import random

import factory
from wbcore.contrib.directory.factories.entries import CompanyFactory
from wbfdm.factories.instruments import InstrumentFactory, InstrumentTypeFactory

from wbportfolio.factories import ModelPortfolioFactory, PortfolioFactory
from wbportfolio.models import InstrumentPortfolioThroughModel
from wbportfolio.models.products import (
    AssetClass,
    InvestmentIndex,
    LegalStructure,
    Liquidy,
    Product,
    TypeOfReturn,
)


class ProductFactory(InstrumentFactory):
    class Meta:
        model = Product

    share_price = 100
    issue_price = factory.LazyAttribute(lambda o: random.randint(1, 1000))
    bank = factory.SubFactory("wbcore.contrib.directory.factories.CompanyFactory")

    type_of_return = factory.Faker("random_element", elements=[x[0] for x in TypeOfReturn.choices()])
    asset_class = factory.Faker("random_element", elements=[x[0] for x in AssetClass.choices()])
    legal_structure = factory.Faker("random_element", elements=[x[0] for x in LegalStructure.choices()])
    investment_index = factory.Faker("random_element", elements=[x[0] for x in InvestmentIndex.choices()])
    liquidity = factory.Faker("random_element", elements=[x[0] for x in Liquidy.choices()])
    risk_scale = factory.LazyAttribute(lambda o: random.randint(1, 7))

    dividend = factory.Faker("random_element", elements=[x[0] for x in Product.Dividend.choices])
    minimum_subscription = factory.LazyAttribute(lambda o: random.randint(100, 200))

    cut_off_time = factory.Faker("time_object")

    external_webpage = factory.Faker("url")
    instrument_type = factory.LazyAttribute(lambda o: InstrumentTypeFactory.create(name="Product", key="product"))

    @factory.post_generation
    def create_initial_portfolio(product, *args, **kwargs):
        if product.id and not product.portfolios.exists():
            portfolio = PortfolioFactory.create()
            InstrumentPortfolioThroughModel.objects.create(instrument=product, portfolio=portfolio)

    # wbportfolio = factory.SubFactory(PortfolioFactory)
    # portfolio_computed = factory.SubFactory(PortfolioFactory)


class WhiteLabelProductFactory(ProductFactory):
    @factory.post_generation
    def white_label_customers(self, create, extracted, **kwargs):
        if not create:
            return

        num_of_wlc = extracted or 1

        for _ in range(num_of_wlc):
            self.white_label_customers.add(CompanyFactory.create())


class IndexProductFactory(ProductFactory):
    pass


class ModelPortfolioWithBaseProductFactory(ModelPortfolioFactory):
    @factory.post_generation
    def create_instrument(self, create, extracted, **kwargs):
        if create:
            instrument = ProductFactory.create()
            InstrumentPortfolioThroughModel.objects.update_or_create(
                instrument=instrument, defaults={"portfolio": self}
            )



from wbcore import serializers as wb_serializers
from wbfdm.serializers import (
    ClassifiableInstrumentRepresentationSerializer,
    ClassificationIsFavoriteZeroHeightRepresentationSerializer,
    ManagedInstrumentRepresentationSerializer,
)
from wbfdm.serializers.instruments.classifications import (
    ClassificationGroupRepresentationSerializer,
)

from wbportfolio.models.portfolio_relationship import (
    InstrumentPortfolioThroughModel,
    PortfolioInstrumentPreferredClassificationThroughModel,
)

from .portfolios import PortfolioRepresentationSerializer


class InstrumentPortfolioThroughModelSerializer(wb_serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    _instrument = ManagedInstrumentRepresentationSerializer(source="instrument")

    class Meta:
        model = InstrumentPortfolioThroughModel
        fields = (
            "id",
            "_portfolio",
            "portfolio",
            "_instrument",
            "instrument",
        )


class InstrumentPreferedClassificationThroughProductModelSerializer(wb_serializers.ModelSerializer):
    _instrument = ClassifiableInstrumentRepresentationSerializer(source="instrument")
    _classification = ClassificationIsFavoriteZeroHeightRepresentationSerializer(
        source="classification", optional_get_parameters={"instrument": "instrument"}
    )
    _classification_group = ClassificationGroupRepresentationSerializer(source="classification_group")

    class Meta:
        model = PortfolioInstrumentPreferredClassificationThroughModel
        dependency_map = {"classification": ["instrument"]}
        fields = (
            "id",
            "portfolio",
            "_instrument",
            "instrument",
            "_classification",
            "classification",
            "classification_group",
            "_classification_group",
            "_additional_resources",
        )



from contextlib import suppress

from rest_framework.reverse import reverse
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig

from wbportfolio.models.portfolio import Portfolio, PortfolioPortfolioThroughModel


class AssetPositionButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return set()

    def get_custom_instance_buttons(self):
        return set()


class AssetPositionPortfolioButtonConfig(AssetPositionButtonConfig):
    def get_custom_buttons(self):
        btns = []
        with suppress(Portfolio.DoesNotExist):
            portfolio = Portfolio.objects.get(id=self.view.kwargs.get("portfolio_id", None))
            btns.append(
                bt.WidgetButton(
                    endpoint=reverse(
                        "wbportfolio:portfolio-contributor-list", args=[portfolio.id], request=self.request
                    ),
                    label="Contributor",
                )
            )
            btns.append(
                bt.WidgetButton(
                    endpoint=(
                        reverse(
                            "wbportfolio:portfolio-distributionchart-list",
                            args=[portfolio.id],
                            request=self.request,
                        )
                    ),
                    label="Distribution Charts",
                )
            )
            btns.append(
                bt.WidgetButton(
                    endpoint=(
                        reverse(
                            "wbportfolio:portfolio-esgaggregation-list",
                            args=[portfolio.id],
                            request=self.request,
                        )
                    ),
                    label="ESG Aggregation",
                )
            )
            cash_management_btns = []

            if portfolio.daily_cashflows.all().exists():
                cash_management_btns.append(
                    bt.WidgetButton(
                        label="Cash Flow",
                        endpoint=reverse(
                            "wbportfolio:portfolio-portfoliocashflow-list",
                            args=[portfolio.id],
                            request=self.request,
                        ),
                    )
                )

            if portfolio.bank_accounts.all().exists():
                bank_account_ids = map(lambda i: str(i), portfolio.bank_accounts.all().values_list("id", flat=True))
                base_url = reverse("wbaccounting:futurecashflow-list", request=self.request)
                cash_flow_url = f"{base_url}?banking_contact={", ".join(bank_account_ids)}"
                cash_management_btns.append(
                    bt.WidgetButton(
                        label="Bank Accounts",
                        endpoint=cash_flow_url,
                    )
                )

            if len(cash_management_btns) > 0:
                btns.append(
                    bt.DropDownButton(
                        label="Cash Management",
                        buttons=cash_management_btns,
                    )
                )
            for rel in PortfolioPortfolioThroughModel.objects.filter(portfolio=portfolio):
                dependency_portfolio = rel.dependency_portfolio
                if dependency_portfolio.assets.exists():
                    btns.append(
                        bt.WidgetButton(
                            endpoint=(
                                reverse(
                                    "wbportfolio:portfolio-asset-list",
                                    args=[dependency_portfolio.id],
                                    request=self.request,
                                )
                            ),
                            label=f"{PortfolioPortfolioThroughModel.Type[rel.type].label} Portfolio",
                        )
                    )
        return set(btns)


class AssetPositionInstrumentButtonConfig(AssetPositionButtonConfig):
    def get_custom_buttons(self):
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            return {
                bt.WidgetButton(
                    endpoint=reverse(
                        "wbportfolio:instrument-assetpositionchart-list", args=[instrument_id], request=self.request
                    ),
                    label="Portfolio Allocation ",
                ),
            }
        return set()


class DistributionChartButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set:
        if portfolio_id := self.view.kwargs.get("portfolio_id", None):
            return {
                bt.WidgetButton(
                    endpoint=(
                        f"{reverse('wbportfolio:portfolio-distributiontable-list', args=[portfolio_id], request=self.request)}"
                        f"?group_by={self.request.GET.get('group_by')}"
                    ),
                    label="Table Form",
                ),
            }
        return set()


class DistributionTableButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set:
        if portfolio_id := self.view.kwargs.get("portfolio_id", None):
            return {
                bt.WidgetButton(
                    endpoint=(
                        f"{reverse('wbportfolio:portfolio-distributionchart-list', args=[portfolio_id], request=self.request)}"
                        f"?group_by={self.request.GET.get('group_by')}"
                    ),
                    label="Chart Form",
                ),
            }
        return set()



from datetime import datetime
from typing import Any, Dict, Optional

from django.db import models
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.io.imports import ImportExportHandler


class RegisterImportHandler(ImportExportHandler):
    MODEL_APP_LABEL = "wbportfolio.Register"

    def _deserialize(self, data: Dict[str, Any]):
        if "opened" in data:
            data["opened"] = datetime.strptime(data["opened"], "%Y-%m-%d").date()
        if residence_id := data.get("residence"):
            data["residence"] = Geography.countries.get(id=residence_id)
        if citizenship_id := data.get("citizenship"):
            data["citizenship"] = Geography.countries.get(id=citizenship_id)
        if outlet_country_id := data.get("outlet_country"):
            data["outlet_country"] = Geography.countries.get(id=outlet_country_id)
        if custodian_country_id := data.get("custodian_country"):
            data["custodian_country"] = Geography.countries.get(id=custodian_country_id)

        outlet_city = data.pop("outlet_city", None)
        if (outlet_country := data.get("outlet_country", None)) and outlet_city:
            if res := outlet_country.lookup_descendants(outlet_city, level=Geography.Level.CITY.value):
                data["outlet_city"] = res

        custodian_city = data.pop("custodian_city", None)
        if (custodian_country := data.get("custodian_country", None)) and custodian_city:
            if res := custodian_country.lookup_descendants(custodian_city, level=Geography.Level.CITY.value):
                data["custodian_city"] = res

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        self.import_source.log += "Get Register Instance."
        self.import_source.log += f"Parameter: Reference={data['register_reference']}"
        if isinstance(data, int):
            return self.model.objects.get(id=data)
        return self.model.objects.filter(register_reference=data["register_reference"]).first()

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        self.import_source.log += "Create Register."
        return self.model.objects.create(import_source=self.import_source, **data)



from contextlib import suppress
from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Section
from wbcore.metadata.configs.display.instance_display import Inline, Layout, Page
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.instance_display.shortcuts import Display
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbportfolio.models import OrderProposal


class OrderProposalDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="portfolio", label="Portfolio") if "portfolio_id" not in self.view.kwargs else None,
                dp.Field(key="trade_date", label="Order Date"),
                dp.Field(key="rebalancing_model", label="Rebalancing Model"),
                dp.Field(key="comment", label="Comment"),
                dp.Field(key="creator", label="Creator"),
                dp.Field(key="approver", label="Approver"),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=OrderProposal.Status.DRAFT.label,
                            value=OrderProposal.Status.DRAFT.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=OrderProposal.Status.PENDING.label,
                            value=OrderProposal.Status.PENDING.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=OrderProposal.Status.APPROVED.label,
                            value=OrderProposal.Status.APPROVED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=OrderProposal.Status.DENIED.label,
                            value=OrderProposal.Status.DENIED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN.value,
                            label=OrderProposal.Status.APPLIED.label,
                            value=OrderProposal.Status.APPLIED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREY.value,
                            label=OrderProposal.Status.EXECUTION.label,
                            value=OrderProposal.Status.EXECUTION.value,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", OrderProposal.Status.DRAFT.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", OrderProposal.Status.PENDING.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", OrderProposal.Status.APPROVED.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN.value},
                            condition=("==", OrderProposal.Status.APPLIED.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREY.value},
                            condition=("==", OrderProposal.Status.EXECUTION.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", OrderProposal.Status.DENIED.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_DARK.value},
                            condition=("==", OrderProposal.Status.FAILED.value),
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        orders_grid_template_areas = [["orders"]]
        orders_grid_template_rows = ["1fr"]
        sections = []
        with suppress(AttributeError, AssertionError):
            op = self.view.get_object()
            if op.execution_status:
                orders_grid_template_areas = [["execution"], ["orders"]]
                orders_grid_template_rows = ["100px", "1fr"]
                sections.append(
                    Section(
                        key="execution",
                        title="Execution",
                        collapsed=False,
                        display=Display(
                            pages=[
                                Page(
                                    layouts={
                                        default(): Layout(
                                            grid_template_areas=[["execution_status_repr", "execution_comment"]],
                                            grid_template_columns=["0.3fr", "0.7fr"],
                                        )
                                    }
                                )
                            ]
                        ),
                    )
                )
        return Display(
            pages=[
                Page(
                    title="Main Information",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["status", "status", "status", "status"],
                                ["trade_date", "total_cash_weight", "min_order_value", "min_weighting"],
                                ["rebalancing_model", "rebalancing_model", "target_portfolio", "target_portfolio"]
                                if self.view.new_mode
                                else ["rebalancing_model", "rebalancing_model", "creator", "approver"],
                                ["comment", "comment", "comment", "comment"],
                            ],
                        ),
                    },
                ),
                Page(
                    title="Orders",
                    layouts={
                        default(): Layout(
                            grid_template_areas=orders_grid_template_areas,
                            grid_template_rows=orders_grid_template_rows,
                            inlines=[Inline(key="orders", endpoint="orders")],
                            sections=sections,
                        ),
                    },
                ),
            ]
        )



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from django.db import connection
from django.db.models import DateField, DecimalField, ExpressionWrapper, F, OuterRef, Subquery
from django.db.models.functions import Greatest
from django.template.loader import get_template
from jinjasql import JinjaSql
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.dataloader.utils import dictfetchall
from wbcrm.models import Account
from wbfdm.models import ClassificationGroup
from wbfdm.models.instruments import Classification, InstrumentPrice

from wbportfolio.models.products import Product

if TYPE_CHECKING:
    from datetime import date
    from typing import Iterable


def get_assets_and_net_new_money_progression(
    from_date: "date", to_date: "date", product_id: int | None = None, account_tree_id: int | None = None
) -> "Iterable[dict]":
    template = get_template("wbportfolio/sql/aum_nnm.sql", using="jinja").template
    query, params = JinjaSql(param_style="format").prepare_query(
        template,
        {
            "from_date": from_date.strftime("%Y-%m-%d"),
            "to_date": to_date.strftime("%Y-%m-%d"),
            "product_id": product_id,
            "account_tree_id": account_tree_id,
        },
    )

    with connection.cursor() as cursor:
        cursor.execute(query, params)
        yield from dictfetchall(cursor)


class ConsolidatedTradeSummary:
    def __init__(
        self,
        queryset,
        start_date: "date",
        end_date: "date",
        pivot: str,
        pivot_label: str,
        classification_group: ClassificationGroup | None = None,
        classification_height: int = 0,
        date_label: str = "date_considered",
    ):
        self.queryset = queryset.filter(trade__isnull=False, status="APPROVED")
        self.start_date = start_date
        self.end_date = end_date
        self.pivot = pivot
        self.pivot_label = pivot_label
        self.queryset = self.queryset.annotate(
            internal_trade=F("trade__marked_as_internal"),
            valid_date=Greatest("trade__transaction_date", "date"),
            date_considered=ExpressionWrapper(F("valid_date") + 1, output_field=DateField()),
            net_value=InstrumentPrice.subquery_closest_value(
                "net_value", date_name="valid_date", instrument_pk_name="product__pk", date_lookup="exact"
            ),
            fx_rate=CurrencyFXRates.get_fx_rates_subquery("valid_date", lookup_expr="exact"),
            aum=ExpressionWrapper(F("fx_rate") * F("net_value") * F("shares"), output_field=DecimalField()),
        )
        self.queryset = Account.annotate_root_account_info(self.queryset)

        if classification_group:
            self.queryset = classification_group.annotate_queryset(self.queryset, classification_height, "product")

        self.queryset = self.queryset.select_related(
            "product",
            "account",
            "product__parent",
        )
        self.df = self._prepare_df(date_label)

    def _prepare_df(self, date_label: str) -> pd.DataFrame:
        columns = ["shares", date_label, "product__id", "aum", "internal_trade"]
        if self.pivot == "classification_id":
            columns.append("classifications")
        else:
            columns.append(self.pivot)
            columns.append(self.pivot_label)

        df = pd.DataFrame(self.queryset.values_list(*columns), columns=columns).rename(columns={date_label: "date"})
        if self.pivot == "classification_id":
            df = (
                df.explode("classifications")
                .rename(columns={"classifications": "classification_id"})
                .replace([np.inf, -np.inf, np.nan], None)
            )
            df["classification_title"] = df.classification_id.map(
                dict(Classification.objects.filter(id__in=df.classification_id.unique()).values_list("id", "name")),
                na_action="ignore",
            )
        df_product = pd.DataFrame(
            Product.objects.annotate(
                net_value_date_start=Subquery(
                    InstrumentPrice.objects.filter(
                        instrument=OuterRef("pk"), calculated=False, date__lte=self.start_date
                    )
                    .order_by("-date")
                    .values("date")[:1]
                ),
                net_value_date_end=Subquery(
                    InstrumentPrice.objects.filter(
                        instrument=OuterRef("pk"), calculated=False, date__lte=self.end_date
                    )
                    .order_by("-date")
                    .values("date")[:1]
                ),
                net_value_start=InstrumentPrice.subquery_closest_value(
                    "net_value",
                    date_name="net_value_date_start",
                    instrument_pk_name="pk",
                    date_lookup="exact",
                    # we get all net value (even estimated) to avoir showing None price on holiday
                ),
                fx_rate_start=CurrencyFXRates.get_fx_rates_subquery(
                    "net_value_date_start", currency="currency", lookup_expr="exact"
                ),
                price_start=ExpressionWrapper(F("net_value_start") * F("fx_rate_start"), output_field=DecimalField()),
                net_value_end=InstrumentPrice.subquery_closest_value(
                    "net_value",
                    date_name="net_value_date_end",
                    instrument_pk_name="pk",
                    date_lookup="exact",
                    # we get all net value (even estimated) to avoir showing None price on holiday
                ),
                fx_rate_end=CurrencyFXRates.get_fx_rates_subquery(
                    "net_value_date_end", currency="currency", lookup_expr="exact"
                ),
                price_end=ExpressionWrapper(F("net_value_end") * F("fx_rate_end"), output_field=DecimalField()),
            ).values("price_start", "price_end", "id"),
            columns=["price_start", "price_end", "id"],
        ).rename(columns={"id": "product__id"})

        df = df.merge(df_product, on="product__id")

        df = df.rename(columns={self.pivot: "id", self.pivot_label: "title"})
        df.loc[df["id"].isnull(), "id"] = -1
        df.loc[df["id"] == -1, "title"] = "N/A"
        df = df.dropna(subset=["id"])
        df[df.columns.difference(["title", "id", "date", "product__id", "internal_trade"])] = df[
            df.columns.difference(["title", "id", "date", "product__id", "internal_trade"])
        ].astype("float")
        return df[df["date"] < self.end_date]

    def get_aum_df(self) -> pd.DataFrame:
        df = self.df.copy().drop(columns=["internal_trade"])
        df.loc[df.date <= self.start_date, "sum_shares_start"] = df.shares
        df.loc[df.date < self.end_date, "sum_shares_end"] = df.shares
        df = df.fillna(0)
        # Prepare dataframe
        df["sum_aum_start"] = df.sum_shares_start * df.price_start

        df["sum_aum_end"] = df.sum_shares_end * df.price_end
        # Sanitize dataframe
        df = df.drop(
            columns=df.columns.difference(["id", "sum_shares_start", "sum_shares_end", "sum_aum_start", "sum_aum_end"])
        )
        df = df.groupby("id").agg(
            {
                "sum_shares_start": "sum",
                "sum_shares_end": "sum",
                "sum_aum_start": "sum",
                "sum_aum_end": "sum",
            }
        )
        df = df.round()
        # Compute statistics
        df["sum_shares_diff"] = df.sum_shares_end - df.sum_shares_start
        df["sum_shares_perf"] = df.sum_shares_end / df.sum_shares_start - 1
        df["sum_aum_diff"] = df.sum_aum_end - df.sum_aum_start
        df["sum_aum_perf"] = df.sum_aum_end / df.sum_aum_start - 1
        return df

    def get_nnm_df(
        self, only_positive: bool = False, only_negative: bool = False, filter_internal_trade: bool = True
    ) -> pd.DataFrame:
        df = self.df.copy()
        if filter_internal_trade:
            df = df[~df["internal_trade"]].drop(columns=["internal_trade"])
        if only_positive:
            df.loc[df["aum"] < 0, "aum"] = 0
        elif only_negative:
            df.loc[df["aum"] > 0, "aum"] = 0
        df = df[((df["date"] > self.start_date) & (df["date"] <= self.end_date))]
        df["date"] = pd.to_datetime(df["date"])
        df["period"] = "sum_nnm_" + df.date.dt.year.astype(str) + "-" + df.date.dt.month.map("{:02}".format)
        df = df[["id", "period", "aum"]].groupby(["id", "period"]).sum().reset_index()
        df = pd.pivot_table(df, values="aum", index="id", columns=["period"])
        df["sum_nnm_total"] = df.sum(axis=1)
        df = df.round().fillna(0).reindex(self.df["id"].unique(), fill_value=0)
        return df

    def get_aum_sparkline(self) -> pd.DataFrame:
        df = self.df[self.df["date"] >= self.start_date].copy()
        df["aum_sparkline"] = None
        if not df.empty:
            df["aum_sparkline"] = df["aum"]
            df = df[["product__id", "id", "date", "aum_sparkline"]].groupby(["product__id", "id", "date"]).sum()
            df = df.reindex(
                pd.MultiIndex.from_product(
                    [
                        df.index.levels[0],
                        df.index.levels[1],
                        pd.date_range(df.index.levels[2].min(), df.index.levels[2].max()),
                    ],  # we downsample to make the chart easier on the frontend
                    names=["product__id", "id", "date"],
                ),
                fill_value=0,
            )
            df["aum_sparkline"] = df.groupby(level=["product__id", "id"])["aum_sparkline"].cumsum()
            df = df.groupby(level=[1, 2]).sum()
            df = df.reindex(
                pd.MultiIndex.from_product(
                    [
                        df.index.levels[0],
                        pd.date_range(df.index.levels[1].min(), df.index.levels[1].max(), freq="W-MON"),
                    ],  # we downsample to make the chart easier on the frontend
                    names=["id", "date"],
                ),
                fill_value=0,
            )
            df = df.reset_index(level=1)[["date", "aum_sparkline"]].apply(tuple, axis=1).rename("aum_sparkline")
            return df.groupby("id").apply(list)
        return pd.DataFrame()

    def get_initial_investment_date_df(self) -> pd.DataFrame:
        return (
            self.df[["date", "id"]]
            .sort_values(by="date")
            .groupby("id")
            .first()
            .rename(columns={"date": "initial_investment_date"})
        )



import pytest

from wbportfolio.factories import (
    ManagerPortfolioRoleFactory,
    ProductPortfolioRoleFactory,
)
from wbportfolio.models import PortfolioRole
from wbportfolio.permissions import is_analyst, is_manager, is_portfolio_manager


@pytest.mark.django_db
class TestPortfolioRoleModel:
    def test_init_product_portfolio_role(self, product_portfolio_role):
        assert product_portfolio_role.id is not None

    def test_init_manager_portfolio_role(self, manager_portfolio_role):
        assert manager_portfolio_role.id is not None

    def test_str(self, product_portfolio_role):
        assert (
            str(product_portfolio_role)
            == f"{product_portfolio_role.role_type} {product_portfolio_role.person.computed_str}"
        )

    # test is_manager

    def test_is_manager_portfolio_role(self, user):
        ManagerPortfolioRoleFactory.create(person=user.profile)
        assert PortfolioRole.is_manager(user.profile)

    def test_is_not_manager_portfolio_role(self, user):
        assert not PortfolioRole.is_manager(user.profile)

    def test_is_manager_portfolio_role_superuser(self, superuser):
        superuser.is_superuser = True
        superuser.save()
        assert PortfolioRole.is_manager(superuser.profile)

    # test is_portfolio_manager

    def test_is_portfolio_manager_portfolio_role(self, user, product):
        ProductPortfolioRoleFactory.create(
            person=user.profile, instrument=product, role_type=PortfolioRole.RoleType.PORTFOLIO_MANAGER
        )
        assert PortfolioRole.is_portfolio_manager(user.profile, product)

    def test_is_portfolio_manager_portfolio_role_manager(self, user, product):
        ManagerPortfolioRoleFactory.create(person=user.profile)
        assert PortfolioRole.is_portfolio_manager(user.profile, product)

    def test_is_not_portfolio_manager_portfolio_role(self, user, product):
        assert not PortfolioRole.is_portfolio_manager(user.profile, product)

    def test_is_portfolio_manager_portfolio_role_superuser(self, superuser, product):
        assert PortfolioRole.is_portfolio_manager(superuser.profile, product)

    # test is_analyst

    def test_is_analyst_portfolio_role(self, user, product):
        ProductPortfolioRoleFactory.create(
            person=user.profile, instrument=product, role_type=PortfolioRole.RoleType.ANALYST
        )
        assert PortfolioRole.is_analyst(user.profile, product)

    def test_is_analyst_portfolio_role_manager(self, user, product):
        ManagerPortfolioRoleFactory.create(person=user.profile)
        assert PortfolioRole.is_analyst(user.profile, product)

    def test_is_not_analyst_portfolio_role(self, user, product):
        assert not PortfolioRole.is_analyst(user.profile, product)

    def test_is_analyst_portfolio_role_superuser(self, superuser, product):
        assert PortfolioRole.is_analyst(superuser.profile, product)

    def test_permissions(self, user):
        class Request:
            def __init__(self, user):
                self.user = user

        request = Request(user)
        assert is_manager(request) == PortfolioRole.is_manager(user.profile)
        assert is_analyst(request) == PortfolioRole.is_analyst(user.profile)
        assert is_portfolio_manager(request) == PortfolioRole.is_portfolio_manager(user.profile)



from decimal import Decimal

import numpy as np
import pandas as pd
from django.db.models import QuerySet
from pandas._libs.tslibs.offsets import BDay
from wbfdm.enums import MarketData
from wbfdm.models import (
    Classification,
    Exchange,
    Instrument,
    InstrumentListThroughModel,
)

from wbportfolio.pms.analytics.utils import fix_quantization_error
from wbportfolio.pms.typing import Portfolio, Position
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Market Capitalization Rebalancing")
class MarketCapitalizationRebalancing(AbstractRebalancingModel):
    TARGET_CURRENCY: str = "USD"
    MIN_WEIGHT: float = 1e-5  # we allow only weight of minimum 0.01%

    def __init__(
        self,
        *args,
        bypass_exchange_check: bool = False,
        ffill_market_cap_limit: int = 5,
        effective_portfolio: PortfolioDTO | None = None,
        **kwargs,
    ):
        super().__init__(*args, effective_portfolio=effective_portfolio, **kwargs)
        self.bypass_exchange_check = bypass_exchange_check
        instruments = self._get_instruments(**kwargs)
        self.market_cap_df = pd.DataFrame(
            instruments.dl.market_data(
                values=[MarketData.MARKET_CAPITALIZATION],
                from_date=self.trade_date - BDay(ffill_market_cap_limit),
                to_date=self.trade_date,
                target_currency=self.TARGET_CURRENCY,
            )
        )
        self.exchange_df = (
            pd.DataFrame(instruments.values_list("id", "exchange"), columns=["id", "exchange"])
            .set_index("id")
            .replace({np.nan: None})
        )
        instrument_ids = list(instruments.values_list("id", flat=True))
        try:
            self.market_cap_df = (
                self.market_cap_df.sort_values(by="valuation_date")
                .groupby("instrument_id")
                .last()["market_capitalization"]
            )
            self.market_cap_df = self.market_cap_df.reindex(instrument_ids)
        except (IndexError, KeyError):
            self.market_cap_df = pd.Series(dtype="float64", index=instrument_ids)

    def _get_instruments(
        self,
        classification_ids: list[int] | None = None,
        instrument_ids: list[int] | None = None,
        instrument_list_id: int | None = None,
    ) -> QuerySet[Instrument]:
        """
        Use the provided kwargs to return a list of instruments as universe.
        - If classifications are given, we returns all the instrument linked to these classifications
        - Or directly from a static list of instrument ids
        - fallback to the last effective portfolio underlying instruments list
        """
        if not instrument_ids:
            instrument_ids = []
        if classification_ids:
            for classification in Classification.objects.filter(id__in=classification_ids):
                for children in classification.get_descendants(include_self=True):
                    for company in children.instruments.all():
                        instrument_ids.append(company.get_primary_quote().id)
        if instrument_list_id:
            instrument_ids.extend(
                list(
                    InstrumentListThroughModel.objects.filter(instrument_list_id=instrument_list_id).values_list(
                        "instrument", flat=True
                    )
                )
            )

        if not instrument_ids:
            instrument_ids = list(self.effective_portfolio.positions_map.keys())

        return (
            Instrument.objects.filter(id__in=instrument_ids)
            .filter(exchange__isnull=False)
            .filter_active_at_date(self.trade_date)
        )

    def is_valid(self) -> bool:
        if not self.market_cap_df.empty:
            df = pd.concat(
                [self.market_cap_df, self.exchange_df], axis=1
            )  # if we are missing any market cap for not-delisted instrument, we consider the rebalancing not valid
            df = df.groupby("exchange", dropna=False)["market_capitalization"].any()
            missing_exchanges = Exchange.objects.filter(id__in=df[~df].index.to_list())
            # if bypass exchange check is true, we do not care whether an exchange is closed we just care if there are at least one exchange open
            if self.bypass_exchange_check:
                return df.any()
            else:
                if missing_exchanges.exists():
                    setattr(
                        self,
                        "_validation_errors",
                        f"Couldn't find any market capitalization for exchanges {', '.join([str(e) for e in missing_exchanges])}",
                    )
                return df.all()
        return False

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        df = self.market_cap_df / self.market_cap_df.dropna().sum()
        df = df[df > self.MIN_WEIGHT]
        df = df / df.sum()
        df = fix_quantization_error(df, 8)
        for underlying_instrument, weighting in df.to_dict().items():
            if np.isnan(weighting):
                weighting = Decimal(0)
            else:
                weighting = round(Decimal(weighting), 8)
            positions.append(
                Position(
                    underlying_instrument=underlying_instrument,
                    weighting=weighting,
                    date=self.trade_date,
                )
            )
        return Portfolio(positions=tuple(positions))



# Generated by Django 4.2.9 on 2024-02-12 11:34

from datetime import date
from decimal import Decimal

import django.contrib.postgres.constraints
import django.contrib.postgres.fields.ranges
import django.db.models.deletion
from django.db import migrations, models
from psycopg.types.range import DateRange


def create_fees(apps, schema_editor):
    Product = apps.get_model("wbportfolio", "Product")
    FeeProductPercentage = apps.get_model("wbportfolio", "FeeProductPercentage")
    for product in Product.objects.all():
        FeeProductPercentage.objects.get_or_create(
            product=product,
            type="BANK",
            timespan=DateRange(date.min, date.max),
            defaults={"percent": product.bank_fees},
        )
        FeeProductPercentage.objects.get_or_create(
            product=product,
            type="MANAGEMENT",
            timespan=DateRange(date.min, date.max),
            defaults={"percent": product.management_fees},
        )
        FeeProductPercentage.objects.get_or_create(
            product=product,
            type="PERFORMANCE",
            timespan=DateRange(date.min, date.max),
            defaults={"percent": product.performance_fees, "vat_deduction": product.performance_fees_vat},
        )


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0054_portfolioinstrumentpreferredclassificationthroughmodel_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="FeeProductPercentage",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "type",
                    models.CharField(
                        choices=[("MANAGEMENT", "Management"), ("PERFORMANCE", "Performance"), ("BANK", "Bank")],
                        default="MANAGEMENT",
                        max_length=16,
                    ),
                ),
                (
                    "vat_deduction",
                    models.DecimalField(
                        decimal_places=6,
                        default=Decimal("0"),
                        help_text="The VAT deducted from this fees percentage",
                        max_digits=6,
                        verbose_name="VAT Deduction",
                    ),
                ),
                ("percent", models.DecimalField(decimal_places=6, max_digits=6)),
                ("timespan", django.contrib.postgres.fields.ranges.DateRangeField()),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fees_percentages",
                        to="wbportfolio.product",
                    ),
                ),
            ],
            options={
                "verbose_name": "Fees Percentage",
                "verbose_name_plural": "Fees Percentage",
            },
        ),
        migrations.RunPython(create_fees),
        migrations.RemoveField(
            model_name="product",
            name="_custom_management_rebates",
        ),
        migrations.RemoveField(
            model_name="product",
            name="_custom_management_rebates_indirect",
        ),
        migrations.RemoveField(
            model_name="product",
            name="_custom_minimum_investment",
        ),
        migrations.RemoveField(
            model_name="product",
            name="_custom_performance_rebates",
        ),
        migrations.RemoveField(
            model_name="product",
            name="_custom_performance_rebates_indirect",
        ),
        migrations.RemoveField(
            model_name="product",
            name="bank_fees",
        ),
        migrations.RemoveField(
            model_name="product",
            name="management_fees",
        ),
        migrations.RemoveField(
            model_name="product",
            name="performance_fees",
        ),
        migrations.RemoveField(
            model_name="product",
            name="performance_fees_vat",
        ),
        migrations.AddConstraint(
            model_name="feeproductpercentage",
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                expressions=[("timespan", "&&"), ("product", "="), ("type", "=")],
                name="exclude_overlapping_product_fees_type",
            ),
        ),
        migrations.RemoveConstraint(
            model_name="portfolioportfoliothroughmodel",
            name="unique_actual",
        ),
        migrations.RemoveConstraint(
            model_name="portfolioportfoliothroughmodel",
            name="unique_benchmark",
        ),
        migrations.AddConstraint(
            model_name="portfolioportfoliothroughmodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(("type", "PRIMARY")), fields=("portfolio", "type"), name="unique_primary"
            ),
        ),
    ]



# Generated by Django 4.2.7 on 2023-12-12 16:27
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0010_classification_classificationgroup_deal_exchange_and_more"),
        ("wbportfolio", "0051_delete_macroreview"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="dailyfundamental",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="fiscalperiod",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="forecast",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="fundamental",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="geographicsegment",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="instrumentlistthroughmodel",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="instrumentprice",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="relatedinstrumentthroughmodel",
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name="cash",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="dailyfundamental",
            name="import_source",
        ),
        migrations.RemoveField(
            model_name="dailyfundamental",
            name="instrument",
        ),
        migrations.RemoveField(
            model_name="equity",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="etf",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="etp",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="fiscalperiod",
            name="import_source",
        ),
        migrations.RemoveField(
            model_name="fiscalperiod",
            name="instrument",
        ),
        migrations.RemoveField(
            model_name="forecast",
            name="import_source",
        ),
        migrations.RemoveField(
            model_name="forecast",
            name="instrument",
        ),
        migrations.RemoveField(
            model_name="fundamental",
            name="import_source",
        ),
        migrations.RemoveField(
            model_name="fundamental",
            name="instrument",
        ),
        migrations.RemoveField(
            model_name="fundamental",
            name="period",
        ),
        migrations.RemoveField(
            model_name="future",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="geographicsegment",
            name="import_source",
        ),
        migrations.RemoveField(
            model_name="geographicsegment",
            name="instrument",
        ),
        migrations.RemoveField(
            model_name="geographicsegment",
            name="period",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="exchanges",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_daily_fundamental",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="imports_allowed",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_forecast",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_fundamental_current",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_fundamental_y0",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_fundamental_y1",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_fundamental_y2",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_fundamental_y3",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_geographic_segment",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_instrument_price",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="last_quarterly_period",
        ),
        migrations.RemoveField(
            model_name="mutualfund",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="option",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="option",
            name="underlying",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="financial_instrument",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="next_quarterly_period",
        ),
        migrations.RemoveField(
            model_name="interestratederivative",
            name="instrument_ptr",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="portfolios",
        ),
        migrations.AddField(
            model_name="portfolio",
            name="instruments",
            field=models.ManyToManyField(
                blank=True,
                help_text="Instruments linked to this instrument",
                related_name="portfolios",
                through="wbportfolio.InstrumentPortfolioThroughModel",
                to="wbfdm.instrument",
                verbose_name="Instruments",
            ),
        ),
        migrations.RenameField(
            model_name="instrument",
            old_name="price_computation",
            new_name="old_price_computation",
        ),
        migrations.AddField(
            model_name="product",
            name="price_computation",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="products",
                to="wbportfolio.pricecomputation",
                verbose_name="Price Computation Method",
            ),
        ),
        migrations.AddField(
            model_name="index",
            name="price_computation",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="indexes",
                to="wbportfolio.pricecomputation",
                verbose_name="Price Computation Method",
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="exchange",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="wbportfolio.exchange",
            ),
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="old_price_computation",
        ),
        migrations.RenameField(model_name="classification", old_name="title", new_name="name"),
        migrations.RenameField(model_name="classificationgroup", old_name="title", new_name="name"),
        migrations.RenameField(model_name="instrument", old_name="title", new_name="name"),
        migrations.RenameField(model_name="instrument", old_name="title_repr", new_name="name_repr"),
        migrations.RenameField(model_name="instrument", old_name="closing_date", new_name="delisted_date"),
        migrations.RenameField(model_name="instrument", old_name="alternative_titles", new_name="alternative_names"),
        migrations.RemoveField(
            model_name="exchangeinstrumentthroughmodel",
            name="exchange",
        ),
        migrations.RemoveField(
            model_name="exchangeinstrumentthroughmodel",
            name="instrument",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="primary_field",
        ),
        migrations.AlterField(
            model_name="instrument",
            name="currency",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="currency.currency",
            ),
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterUniqueTogether(
                    name="adjustment",
                    unique_together={("date", "instrument")},
                ),
                migrations.AddIndex(
                    model_name="adjustment",
                    index=models.Index(fields=["instrument", "-date"], name="instrument_adjustment_idx"),
                ),
                migrations.AddIndex(
                    model_name="assetposition",
                    index=models.Index(
                        fields=["date", "underlying_instrument", "portfolio"], name="wbportfolio_date_0a7c0a_idx"
                    ),
                ),
                migrations.AddIndex(
                    model_name="assetposition",
                    index=models.Index(fields=["date", "underlying_instrument"], name="wbportfolio_date_da82d2_idx"),
                ),
                migrations.AddIndex(
                    model_name="transaction",
                    index=models.Index(
                        fields=["underlying_instrument", "transaction_date"], name="wbportfolio_underly_391e3e_idx"
                    ),
                ),
                migrations.RemoveField(
                    model_name="classification",
                    name="group",
                ),
                migrations.RemoveField(
                    model_name="classification",
                    name="parent",
                ),
                migrations.RemoveField(
                    model_name="deal",
                    name="equity",
                ),
                migrations.RemoveField(
                    model_name="deal",
                    name="import_source",
                ),
                migrations.RemoveField(
                    model_name="deal",
                    name="investors",
                ),
                migrations.RemoveField(
                    model_name="exchange",
                    name="city",
                ),
                migrations.RemoveField(
                    model_name="exchange",
                    name="country",
                ),
                migrations.RemoveField(
                    model_name="instrument",
                    name="classifications",
                ),
                migrations.RemoveField(
                    model_name="instrument",
                    name="country",
                ),
                migrations.RemoveField(
                    model_name="instrument",
                    name="currency",
                ),
                migrations.RemoveField(
                    model_name="instrument",
                    name="headquarter_city",
                ),
                migrations.RemoveField(
                    model_name="instrument",
                    name="import_source",
                ),
                migrations.RemoveField(
                    model_name="instrument",
                    name="related_instruments",
                ),
                migrations.RemoveField(
                    model_name="instrument",
                    name="tags",
                ),
                migrations.RemoveField(
                    model_name="instrumentclassificationrelatedinstrument",
                    name="classified_instrument",
                ),
                migrations.RemoveField(
                    model_name="instrumentclassificationrelatedinstrument",
                    name="related_instrument",
                ),
                migrations.RemoveField(
                    model_name="instrumentclassificationthroughmodel",
                    name="classification",
                ),
                migrations.RemoveField(
                    model_name="instrumentclassificationthroughmodel",
                    name="instrument",
                ),
                migrations.RemoveField(
                    model_name="instrumentclassificationthroughmodel",
                    name="related_instruments",
                ),
                migrations.RemoveField(
                    model_name="instrumentclassificationthroughmodel",
                    name="tags",
                ),
                migrations.RemoveField(
                    model_name="instrumentfavoritegroup",
                    name="instruments",
                ),
                migrations.RemoveField(
                    model_name="instrumentfavoritegroup",
                    name="owner",
                ),
                migrations.RemoveField(
                    model_name="instrumentlistthroughmodel",
                    name="import_source",
                ),
                migrations.RemoveField(
                    model_name="instrumentlistthroughmodel",
                    name="instrument",
                ),
                migrations.RemoveField(
                    model_name="instrumentlistthroughmodel",
                    name="instrument_list",
                ),
                migrations.RemoveField(
                    model_name="instrumentprice",
                    name="currency_fx_rate_to_usd",
                ),
                migrations.RemoveField(
                    model_name="instrumentprice",
                    name="import_source",
                ),
                migrations.RemoveField(
                    model_name="instrumentprice",
                    name="instrument",
                ),
                migrations.RemoveField(
                    model_name="instrumentrequest",
                    name="created_instrument",
                ),
                migrations.RemoveField(
                    model_name="instrumentrequest",
                    name="handler",
                ),
                migrations.RemoveField(
                    model_name="instrumentrequest",
                    name="requester",
                ),
                migrations.RemoveField(
                    model_name="relatedinstrumentthroughmodel",
                    name="instrument",
                ),
                migrations.RemoveField(
                    model_name="relatedinstrumentthroughmodel",
                    name="related_instrument",
                ),
                migrations.AddField(
                    model_name="adjustment",
                    name="instrument",
                    field=models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="pms_adjustments",
                        to="wbfdm.instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="assetposition",
                    name="exchange",
                    field=models.ForeignKey(
                        blank=True,
                        help_text="The exchange where this asset is.",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="assets",
                        to="wbfdm.exchange",
                        verbose_name="Exchange",
                    ),
                ),
                migrations.AddField(
                    model_name="assetposition",
                    name="underlying_instrument",
                    field=models.ForeignKey(
                        help_text="The instrument that is this asset.",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="assets",
                        to="wbfdm.instrument",
                        verbose_name="Underlying Instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="assetposition",
                    name="underlying_instrument_price",
                    field=models.ForeignKey(
                        blank=True,
                        help_text="The instrument price that is this asset.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="assets",
                        to="wbfdm.instrumentprice",
                        verbose_name="Underlying Instrument Price",
                    ),
                ),
                migrations.AddField(
                    model_name="instrumentportfoliothroughmodel",
                    name="instrument",
                    field=models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="through_portfolios",
                        to="wbfdm.instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="index",
                    name="instrument_ptr",
                    field=models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="wbfdm.instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="product",
                    name="instrument_ptr",
                    field=models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="wbfdm.instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="productgroup",
                    name="instrument_ptr",
                    field=models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="wbfdm.instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="portfolio",
                    name="prefered_instrument_classifications",
                    field=models.ManyToManyField(
                        blank=True,
                        related_name="prefered_portfolio_classifications",
                        through="wbportfolio.PortfolioInstrumentPreferedClassificationThroughModel",
                        to="wbfdm.instrument",
                        verbose_name="The Prefered classification per instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="portfolioinstrumentpreferedclassificationthroughmodel",
                    name="classification",
                    field=models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="prefered_classification_throughs",
                        to="wbfdm.classification",
                    ),
                ),
                migrations.AddField(
                    model_name="portfolioinstrumentpreferedclassificationthroughmodel",
                    name="classification_group",
                    field=models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="prefered_classification_group_throughs",
                        to="wbfdm.classificationgroup",
                    ),
                ),
                migrations.AddField(
                    model_name="portfolioinstrumentpreferedclassificationthroughmodel",
                    name="instrument",
                    field=models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="prefered_classification_portfolio",
                        to="wbfdm.instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="portfoliorole",
                    name="instrument",
                    field=models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="portfolio_roles",
                        to="wbfdm.instrument",
                        verbose_name="Instrument",
                    ),
                ),
                migrations.AddField(
                    model_name="transaction",
                    name="underlying_instrument",
                    field=models.ForeignKey(
                        help_text="The instrument that is this transaction.",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="transactions",
                        to="wbfdm.instrument",
                        verbose_name="Underlying Instrument",
                    ),
                ),
            ]
        ),
        migrations.DeleteModel(
            name="Bond",
        ),
        migrations.DeleteModel(
            name="Cash",
        ),
        migrations.DeleteModel(
            name="Equity",
        ),
        migrations.DeleteModel(
            name="ETF",
        ),
        migrations.DeleteModel(
            name="ETP",
        ),
        migrations.DeleteModel(
            name="DailyFundamental",
        ),
        migrations.DeleteModel(
            name="FiscalPeriod",
        ),
        migrations.DeleteModel(
            name="Forecast",
        ),
        migrations.DeleteModel(
            name="Fundamental",
        ),
        migrations.DeleteModel(
            name="Future",
        ),
        migrations.DeleteModel(
            name="GeographicSegment",
        ),
        migrations.DeleteModel(
            name="InterestRateDerivative",
        ),
        migrations.DeleteModel(
            name="MutualFund",
        ),
        migrations.DeleteModel(
            name="Option",
        ),
        migrations.DeleteModel(
            name="ExchangeInstrumentThroughModel",
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Exchange",
                )
            ],
            database_operations=[migrations.AlterModelTable(name="Exchange", table="wbfdm_exchange")],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Classification",
                )
            ],
            database_operations=[migrations.AlterModelTable(name="Classification", table="wbfdm_classification")],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="ClassificationGroup",
                )
            ],
            database_operations=[
                migrations.AlterModelTable(name="ClassificationGroup", table="wbfdm_classificationgroup")
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Instrument",
                )
            ],
            database_operations=[migrations.AlterModelTable(name="Instrument", table="wbfdm_instrument")],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="InstrumentClassificationRelatedInstrument",
                )
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="InstrumentClassificationRelatedInstrument",
                    table="wbfdm_instrumentclassificationrelatedinstrument",
                )
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="InstrumentClassificationThroughModel",
                )
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="InstrumentClassificationThroughModel", table="wbfdm_instrumentclassificationthroughmodel"
                )
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="InstrumentFavoriteGroup",
                )
            ],
            database_operations=[
                migrations.AlterModelTable(name="InstrumentFavoriteGroup", table="wbfdm_instrumentfavoritegroup")
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="InstrumentList",
                )
            ],
            database_operations=[migrations.AlterModelTable(name="InstrumentList", table="wbfdm_instrumentlist")],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="InstrumentListThroughModel",
                )
            ],
            database_operations=[
                migrations.AlterModelTable(name="InstrumentListThroughModel", table="wbfdm_instrumentlistthroughmodel")
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="InstrumentPrice",
                )
            ],
            database_operations=[migrations.AlterModelTable(name="InstrumentPrice", table="wbfdm_instrumentprice")],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="InstrumentRequest",
                )
            ],
            database_operations=[
                migrations.AlterModelTable(name="InstrumentRequest", table="wbfdm_instrumentrequest")
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="RelatedInstrumentThroughModel",
                )
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="RelatedInstrumentThroughModel", table="wbfdm_relatedinstrumentthroughmodel"
                )
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Deal",
                ),
            ],
            database_operations=[migrations.AlterModelTable(name="Deal", table="wbfdm_deal")],
        ),
    ]



from io import BytesIO

import pandas as pd

from wbportfolio.models import Fees


def parse(import_source):
    data = list()
    df = pd.DataFrame()
    try:
        df = pd.read_excel(
            BytesIO(import_source.file.read()),
            engine="openpyxl",
            sheet_name="FEE_CONSOLIDATION",
            header=[0, 2],
            skiprows=1,
        ).dropna(axis=1)
    except (ValueError, IndexError):
        pass

    if not df.empty:
        for isin in df.columns.unique(level=0):
            product_fees_df = df.loc[:, isin]

            product_fees_df = product_fees_df.rename(
                columns={
                    "TRADE DATE": "fee_date",
                    "INDEX SPONSOR FEE": "total_value",
                    "BUSINESS EVENT": "transaction_subtype",
                }
            )

            product_fees_df.loc[:, "fee_date"] = pd.to_datetime(product_fees_df.loc[:, "fee_date"], dayfirst=True)

            management_fees_eur = product_fees_df[
                product_fees_df.loc[:, "transaction_subtype"] == "Management Fee EUR"
            ]
            performance_fees_eur = product_fees_df[
                product_fees_df.loc[:, "transaction_subtype"] == "Performance Fee EUR"
            ]
            management_fees_eur["transaction_subtype"] = Fees.Type.MANAGEMENT
            performance_fees_eur["transaction_subtype"] = Fees.Type.PERFORMANCE

            concat_list = []
            if not management_fees_eur.empty:
                concat_list.append(management_fees_eur)
            if not performance_fees_eur.empty:
                concat_list.append(performance_fees_eur)
            product_fees_df = pd.concat(
                concat_list,
                axis=0,
            )
            product_fees_df = product_fees_df.where(pd.notnull(product_fees_df), None)
            for fees in product_fees_df.to_dict("records"):
                base_data = {
                    "product": {"isin": isin},
                    "fee_date": fees["fee_date"].strftime("%Y-%m-%d"),
                    "calculated": False,
                }

                total_value_gross = fees.get("total_value_gross", fees["total_value"])
                if total_value_gross != 0 or fees["total_value"] != 0:
                    data.append(
                        {
                            "transaction_subtype": fees["transaction_subtype"],
                            "total_value": fees["total_value"],
                            "total_value_gross": total_value_gross,
                            **base_data,
                        }
                    )

    return {"data": data}



# Generated by Django 4.2.9 on 2024-02-07 15:58
from contextlib import suppress

import django.db.models.deletion
from django.db import migrations, models


def assign_hedged_currency(apps, schema_editor):
    Product = apps.get_model("wbportfolio", "Product")
    InstrumentPortfolioThroughModel = apps.get_model("wbportfolio", "InstrumentPortfolioThroughModel")

    for product in Product.objects.filter(index_hedged_currency__isnull=False).distinct():
        for rel in InstrumentPortfolioThroughModel.objects.filter(instrument_id=product.instrument_ptr_id):
            portfolio = rel.portfolio
            portfolio.hedged_currency = product.index_hedged_currency
            portfolio.save()


def modify_portfolio(apps, schema_editor):
    Portfolio = apps.get_model("wbportfolio", "Portfolio")
    PortfolioPortfolioThroughModel = apps.get_model("wbportfolio", "PortfolioPortfolioThroughModel")
    InstrumentPortfolioThroughModel = apps.get_model("wbportfolio", "InstrumentPortfolioThroughModel")
    TradeProposal = apps.get_model("wbportfolio", "TradeProposal")
    Transaction = apps.get_model("wbportfolio", "Transaction")
    Instrument = apps.get_model("wbfdm", "Instrument")
    Portfolio.objects.filter(portfolio_type="MODEL").update(is_manageable=True)
    PortfolioPortfolioThroughModel.objects.filter(is_primary=True).update(type="PRIMARY")
    PortfolioPortfolioThroughModel.objects.filter(is_theme=True).delete()
    for instrument in Instrument.objects.filter(portfolios__isnull=False).distinct():
        with suppress(InstrumentPortfolioThroughModel.DoesNotExist):
            assert_portfolio = InstrumentPortfolioThroughModel.objects.get(
                instrument=instrument, asset_portfolio=True
            ).portfolio
            primary_portfolio_rel = InstrumentPortfolioThroughModel.objects.get(
                instrument=instrument, primary_portfolio=True
            )
            if assert_portfolio != primary_portfolio_rel.portfolio:
                try:
                    primary_relationship = PortfolioPortfolioThroughModel.objects.get(
                        portfolio=assert_portfolio, type="PRIMARY"
                    )
                except PortfolioPortfolioThroughModel.DoesNotExist:
                    primary_relationship = PortfolioPortfolioThroughModel.objects.create(
                        portfolio=assert_portfolio,
                        dependency_portfolio=primary_portfolio_rel.portfolio,
                        type="PRIMARY",
                    )
                if primary_portfolio_rel.portfolio != primary_relationship.dependency_portfolio:
                    if not primary_portfolio_rel.portfolio.assets.exists():
                        pass
                    else:
                        last_date = primary_portfolio_rel.portfolio.assets.latest("date").date
                        if (
                            primary_portfolio_rel.portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(
                                date=last_date
                            )
                        ).count() == 1:
                            primary_relationship.dependency_portfolio = primary_portfolio_rel.portfolio
                            primary_relationship.save()
                        else:
                            raise ValueError("cannot be here")
                Transaction.objects.filter(portfolio=primary_portfolio_rel.portfolio).update(
                    portfolio=assert_portfolio
                )
                primary_portfolio_rel.delete()
        if InstrumentPortfolioThroughModel.objects.filter(instrument=instrument).count() > 1:
            InstrumentPortfolioThroughModel.objects.filter(
                instrument=instrument, primary_portfolio=False, asset_portfolio=False
            ).delete()
    for p in Portfolio.objects.filter(assets__isnull=True, transactions__isnull=True):
        TradeProposal.objects.filter(model_portfolio=p).update(model_portfolio=models.F("portfolio"))
        p.delete()

    for rel in InstrumentPortfolioThroughModel.objects.filter(model_portfolio=True):
        with suppress(InstrumentPortfolioThroughModel.DoesNotExist):
            asset_portfolio = InstrumentPortfolioThroughModel.objects.get(
                instrument=rel.instrument, asset_portfolio=True
            ).portfolio
            if assert_portfolio != rel.portfolio:
                PortfolioPortfolioThroughModel.objects.update_or_create(
                    portfolio=asset_portfolio, dependency_portfolio=rel.portfolio, defaults={"type": "MODEL"}
                )
    # Cleanup


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0012_instrumentprice_created_instrumentprice_modified"),
        ("currency", "0001_initial"),
        ("wbportfolio", "0053_remove_product_group"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="PortfolioInstrumentPreferedClassificationThroughModel",
            new_name="PortfolioInstrumentPreferredClassificationThroughModel",
        ),
        migrations.RenameField(
            model_name="portfolio",
            old_name="prefered_instrument_classifications",
            new_name="preferred_instrument_classifications",
        ),
        migrations.AddField(
            model_name="portfolio",
            name="hedged_currency",
            field=models.ForeignKey(
                blank=True,
                help_text="The hedged currency of the portfolio.",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="hedged_portfolios",
                to="currency.currency",
                verbose_name="Hedged Currency",
            ),
        ),
        migrations.AddField(
            model_name="portfolio",
            name="is_manageable",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="portfolioportfoliothroughmodel",
            name="type",
            field=models.CharField(
                choices=[("PRIMARY", "Primary"), ("MODEL", "Model"), ("BENCHMARK", "Benchmark")],
                default="PRIMARY",
                verbose_name="Type",
            ),
        ),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL IMMEDIATE;",
        ),
        migrations.RunPython(assign_hedged_currency),
        migrations.RunPython(modify_portfolio),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL DEFERRED;",
        ),
        migrations.RemoveField(
            model_name="instrumentportfoliothroughmodel",
            name="base_instrument",
        ),
        migrations.RemoveField(
            model_name="instrumentportfoliothroughmodel",
            name="model_portfolio",
        ),
        migrations.RemoveField(
            model_name="portfolio",
            name="portfolio_type",
        ),
        migrations.RemoveField(
            model_name="portfolioportfoliothroughmodel",
            name="is_primary",
        ),
        migrations.RemoveField(
            model_name="portfolioportfoliothroughmodel",
            name="is_theme",
        ),
        migrations.RemoveField(
            model_name="product",
            name="index_hedged_currency",
        ),
        migrations.RemoveField(
            model_name="instrumentportfoliothroughmodel",
            name="asset_portfolio",
        ),
        migrations.RemoveField(
            model_name="instrumentportfoliothroughmodel",
            name="primary_portfolio",
        ),
        migrations.AlterField(
            model_name="instrumentportfoliothroughmodel",
            name="instrument",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, related_name="through_portfolios", to="wbfdm.instrument"
            ),
        ),
        migrations.AlterField(
            model_name="portfolio",
            name="instruments",
            field=models.ManyToManyField(
                blank=True,
                help_text="Instruments linked to this instrument",
                related_name="portfolios",
                through="wbportfolio.InstrumentPortfolioThroughModel",
                to="wbfdm.instrument",
                verbose_name="Instruments",
            ),
        ),
        migrations.AlterField(
            model_name="portfolio",
            name="preferred_instrument_classifications",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True)),
                related_name="preferred_portfolio_classifications",
                through="wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel",
                to="wbfdm.instrument",
                verbose_name="The Preferred classification per instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioinstrumentpreferredclassificationthroughmodel",
            name="classification",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="preferred_classification_throughs",
                to="wbfdm.classification",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioinstrumentpreferredclassificationthroughmodel",
            name="classification_group",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="preferred_classification_group_throughs",
                to="wbfdm.classificationgroup",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioinstrumentpreferredclassificationthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="preferred_classification_portfolio",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioinstrumentpreferredclassificationthroughmodel",
            name="portfolio",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="preferred_classification_instruments",
                to="wbportfolio.portfolio",
            ),
        ),
        migrations.AddConstraint(
            model_name="portfolioportfoliothroughmodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(("type", "ACTUAL")), fields=("portfolio", "type"), name="unique_actual"
            ),
        ),
        migrations.AddConstraint(
            model_name="portfolioportfoliothroughmodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(("type", "MODEL")), fields=("portfolio", "type"), name="unique_model"
            ),
        ),
        migrations.AddConstraint(
            model_name="portfolioportfoliothroughmodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(("type", "BENCHMARK")), fields=("portfolio", "type"), name="unique_benchmark"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrumentportfoliothroughmodel",
            constraint=models.UniqueConstraint(fields=("instrument",), name="unique_instrument"),
        ),
        migrations.AddConstraint(
            model_name="instrumentportfoliothroughmodel",
            constraint=models.UniqueConstraint(
                fields=("instrument", "portfolio"), name="unique_portfolio_relationship"
            ),
        ),
    ]



from datetime import date

from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.models import BankingContact
from wbfdm.serializers import InstrumentRepresentationSerializer

from wbportfolio.models import Portfolio, PortfolioPortfolioThroughModel
from wbportfolio.serializers.rebalancing import RebalancerRepresentationSerializer


class PortfolioRepresentationSerializer(wb_serializers.RepresentationSerializer):
    def get_filter_params(self, request):
        return {"is_tracked": True}

    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:portfolio-detail")
    _detail_preview = wb_serializers.HyperlinkField(reverse_name="wbportfolio:portfolio-detail")

    class Meta:
        model = Portfolio
        fields = ("id", "name", "_detail", "_detail_preview")


class PortfolioModelSerializer(wb_serializers.ModelSerializer):
    _currency = CurrencyRepresentationSerializer(source="currency")
    _hedged_currency = CurrencyRepresentationSerializer(source="hedged_currency")
    _depends_on = PortfolioRepresentationSerializer(source="depends_on", many=True)
    automatic_rebalancer = wb_serializers.PrimaryKeyRelatedField(read_only=True)
    _automatic_rebalancer = RebalancerRepresentationSerializer(source="automatic_rebalancer")
    _instruments = InstrumentRepresentationSerializer(source="instruments", many=True)

    create_index = wb_serializers.BooleanField(write_only=True, default=False)

    last_asset_under_management_usd = wb_serializers.FloatField(read_only=True)
    last_positions = wb_serializers.FloatField(read_only=True)
    last_order_proposal_date = wb_serializers.DateField(read_only=True)
    next_expected_order_proposal_date = wb_serializers.SerializerMethodField(read_only=True)

    def get_next_expected_order_proposal_date(self, obj):
        if (automatic_rebalancer := getattr(obj, "automatic_rebalancer", None)) and (
            _d := automatic_rebalancer.get_next_rebalancing_date(date.today())
        ):
            return _d.strftime("%Y-%m-%d")

    def create(self, validated_data):
        create_index = validated_data.pop("create_index", False)
        obj = super().create(validated_data)
        if create_index:
            obj.get_or_create_index()
        return obj

    @wb_serializers.register_only_instance_resource()
    def cash_management(self, instance, request, user, **kwargs):
        additional_resources = dict()
        if instance.daily_cashflows.exists():
            additional_resources["daily_cashflows"] = reverse(
                "wbportfolio:portfolio-portfoliocashflow-list", args=[instance.id], request=request
            )
        b = BankingContact.objects.filter(wbportfolio_products__id__in=instance.instruments.all().values("id"))
        if b.exists():
            base_url = reverse("wbaccounting:futurecashflow-list", request=request)
            additional_resources["cash_flow"] = (
                f'{base_url}?banking_contact={",".join([str(i) for i in b.distinct("id").values_list("id", flat=True)])}'
            )

        return additional_resources

    @wb_serializers.register_only_instance_resource()
    def additional_resources(self, instance, request, user, **kwargs):
        additional_resources = dict()
        additional_resources["distribution_chart"] = reverse(
            "wbportfolio:portfolio-distributionchart-list", args=[instance.id], request=request
        )
        additional_resources["distribution_table"] = reverse(
            "wbportfolio:portfolio-distributiontable-list", args=[instance.id], request=request
        )
        if instance.assets.exists():
            additional_resources["assets"] = reverse(
                "wbportfolio:portfolio-asset-list", args=[instance.id], request=request
            )
            additional_resources["contributor"] = reverse(
                "wbportfolio:portfolio-contributor-list",
                args=[instance.id],
                request=request,
            )
        if user.profile.is_internal:
            additional_resources["instruments_list"] = reverse(
                "wbportfolio:portfolio-instrument-list",
                args=[instance.id],
                request=request,
            )

        additional_resources["dependencyportfolios"] = reverse(
            "wbportfolio:portfolio-dependencyportfolio-list",
            args=[instance.id],
            request=request,
        )

        additional_resources["preferredclassification"] = reverse(
            "wbportfolio:portfolio-preferredclassification-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["modelcomposition"] = reverse(
            "wbportfolio:portfolio-modelcompositionpandas-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["order_proposals"] = reverse(
            "wbportfolio:portfolio-orderproposal-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["rebalance"] = reverse(
            "wbportfolio:portfolio-rebalance",
            args=[instance.id],
            request=request,
        )
        if not getattr(instance, "automatic_rebalancer", None):
            additional_resources["add_automatic_rebalancer"] = reverse(
                "wbportfolio:portfolio-attachrebalancer",
                args=[instance.id],
                request=request,
            )
        if user.is_superuser and instance.is_lookthrough:
            additional_resources["recompute_lookthrough"] = reverse(
                "wbportfolio:portfolio-recomputelookthrough",
                args=[instance.id],
                request=request,
            )
        additional_resources["treegraphchart"] = reverse(
            "wbportfolio:portfolio-treegraphchart-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["topdowncomposition"] = reverse(
            "wbportfolio:portfolio-topdowncomposition-list",
            args=[instance.id],
            request=request,
        )
        return additional_resources

    class Meta:
        model = Portfolio
        fields = (
            "id",
            "name",
            "updated_at",
            "_depends_on",
            "depends_on",
            "_instruments",
            "instruments",
            "_currency",
            "currency",
            "_hedged_currency",
            "hedged_currency",
            "automatic_rebalancer",
            "_automatic_rebalancer",
            "invested_timespan",
            "is_manageable",
            "is_tracked",
            "only_weighting",
            "is_lookthrough",
            "is_composition",
            "_additional_resources",
            "create_index",
            "initial_position_date",
            "last_position_date",
            "last_asset_under_management_usd",
            "last_positions",
            "last_order_proposal_date",
            "next_expected_order_proposal_date",
        )


class PortfolioPortfolioThroughModelSerializer(wb_serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    _dependency_portfolio = PortfolioRepresentationSerializer(source="dependency_portfolio")

    class Meta:
        model = PortfolioPortfolioThroughModel
        fields = (
            "id",
            "_portfolio",
            "portfolio",
            "_dependency_portfolio",
            "dependency_portfolio",
            "type",
        )



from datetime import date

from celery import shared_task
from django.db.models import Exists, OuterRef
from tqdm import tqdm
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Company
from wbcrm.models import Account

from .models import CompanyPortfolioData, Updater


@shared_task(queue="portfolio")
def update_all_portfolio_data(val_date: date | None = None):
    if not val_date:
        val_date = CurrencyFXRates.objects.latest("date").date
    updater = Updater(val_date)
    qs = Company.objects.annotate(
        has_account=Exists(Account.objects.filter(owner=OuterRef("pk"))),
        has_portfolio_data=Exists(CompanyPortfolioData.objects.filter(company=OuterRef("pk"))),
    )
    for company in tqdm(qs, total=qs.count()):
        updater.update_company_data(company)



from django.contrib import admin

from wbportfolio.models import PortfolioRole


class PortfolioRoleInline(admin.TabularInline):
    model = PortfolioRole
    extra = 0

    autocomplete_fields = ["instrument", "person"]


@admin.register(PortfolioRole)
class PortfolioRoleAdmin(admin.ModelAdmin):
    list_display = ["role_type", "instrument", "person", "start", "end"]

    autocomplete_fields = [
        "instrument",
        "person",
    ]



# Generated by Django 5.0.4 on 2024-05-01 11:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0059_fees_unique_fees"),
    ]

    operations = [
        migrations.AlterField(
            model_name="portfolioportfoliothroughmodel",
            name="type",
            field=models.CharField(
                choices=[("PRIMARY", "Primary"), ("MODEL", "Model"), ("BENCHMARK", "Benchmark"), ("INDEX", "Index")],
                default="PRIMARY",
                verbose_name="Type",
            ),
        ),
    ]



import math
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, Optional

from django.db import models
from wbcore.contrib.currency.import_export.handlers import CurrencyImportHandler
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler
from wbfdm.import_export.handlers.instrument import InstrumentImportHandler
from wbfdm.models import InstrumentType

from wbportfolio.models.portfolio import Portfolio
from wbportfolio.models.products import update_outstanding_shares_as_task
from wbportfolio.utils import string_matching

from .register import RegisterImportHandler


class TradeImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbportfolio.Trade"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)
        self.register_handler = RegisterImportHandler(self.import_source)
        self.currency_handler = CurrencyImportHandler(self.import_source)

    def _data_changed(self, _object, change_data: Dict[str, Any], initial_data: Dict[str, Any], **kwargs):
        if (new_register := change_data.get("register")) and (current_register := _object.register):
            # we replace the register only if the new one gives us more information
            if new_register.register_reference == current_register.global_register_reference:
                del change_data["register"]
        return super()._data_changed(_object, change_data, initial_data, **kwargs)

    def _deserialize(self, data: Dict[str, Any]):
        from wbportfolio.models import Product

        if underlying_instrument := data.get("underlying_instrument", None):
            data["underlying_instrument"] = self.instrument_handler.process_object(
                underlying_instrument, only_security=False, read_only=True
            )[0]

        if external_id_alternative := data.get("external_id_alternative", None):
            data["external_id_alternative"] = str(external_id_alternative)
        if transaction_date_str := data.get("transaction_date", None):
            data["transaction_date"] = datetime.strptime(transaction_date_str, "%Y-%m-%d").date()
        if value_date_str := data.get("value_date", None):
            data["value_date"] = datetime.strptime(value_date_str, "%Y-%m-%d").date()
        if book_date_str := data.get("book_date", None):
            data["book_date"] = datetime.strptime(book_date_str, "%Y-%m-%d").date()
        data["portfolio"] = Portfolio._get_or_create_portfolio(
            self.instrument_handler, data.get("portfolio", data["underlying_instrument"])
        )

        if currency_data := data.get("currency", None):
            data["currency"] = self.currency_handler.process_object(currency_data, read_only=True)[0]

        if register_data := data.get("register", None):
            data["register"] = self.register_handler.process_object(register_data)[0]

        data["marked_for_deletion"] = data.get("marked_for_deletion", False)
        if underlying_instrument := data.get("underlying_instrument"):
            if nominal := data.pop("nominal", None):
                try:
                    product = Product.objects.get(id=underlying_instrument.id)
                    data["shares"] = nominal / product.share_price
                except Product.DoesNotExist:
                    raise DeserializationError(
                        "We cannot compute the number of shares from the nominal value as we cannot find the product share price."
                    )
        else:
            raise DeserializationError("We couldn't find a valid underlying instrument this row.")

        for field in self.model._meta.get_fields():
            if not (value := data.get(field.name, None)) is None and isinstance(field, models.DecimalField):
                q = (
                    1 / (math.pow(10, 4))
                )  # we need that convertion mechanism otherwise there is floating point approximation error while casting to decimal and get_instance does not work as expected
                data[field.name] = Decimal(value).quantize(Decimal(str(q)))
        if (target_weight := data.pop("target_weight", None)) is not None:
            data["_target_weight"] = target_weight

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        if "transaction_date" not in data:  # we might get only book date and not transaction date
            data["transaction_date"] = data["book_date"]
        return self.model.objects.create(**data, import_source=self.import_source)

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        self.import_source.log += "\nGet Trade Instance."
        if transaction_date := data.get("transaction_date"):
            dates_lookup = {"transaction_date": transaction_date}
        elif book_date := data.get("book_date"):
            dates_lookup = {"book_date": book_date}
        else:
            raise DeserializationError("date lookup is missing from data")
        self.import_source.log += f"\nParameter: Product={data['underlying_instrument']} Trade-Date={transaction_date} Shares={data.get('shares')} Weighting={data.get('weighting')}"

        if history.exists():
            queryset = history
        else:
            queryset = self.model.objects.filter(marked_for_deletion=False).exclude(id__in=self.processed_ids)

        queryset = queryset.filter(
            models.Q(underlying_instrument=data["underlying_instrument"]) & models.Q(**dates_lookup)
        )
        if "shares" in data:
            queryset = queryset.filter(shares=data["shares"])
        if _id := data.get("id", None):
            self.import_source.log += f"ID {_id} provided -> Load CustomerTrade"
            return self.model.objects.get(id=_id)
        # We need to check for external identifiers
        if external_id := data.get("external_id"):
            self.import_source.log += f"\nExternal Identifier used: {external_id}"
            external_id_queryset = queryset.filter(external_id=external_id)
            if external_id_queryset.count() == 1:
                self.import_source.log += f"External ID {external_id} provided -> Load CustomerTrade"
                return external_id_queryset.first()
        if portfolio := data.get("portfolio", None):
            queryset = queryset.filter(portfolio=portfolio)
        if queryset.exists():
            if bank := data.get("bank"):
                self.import_source.log += (
                    f"\n{queryset.count()} Trades found. The bank will tried to be matched against {bank}"
                )
                if queryset.filter(bank=bank).count() >= 1:  # exact match
                    queryset = queryset.filter(bank=bank)
                else:
                    best_result = string_matching(bank, queryset.values_list("bank", flat=True))
                    if best_result[1] >= 80:
                        possible_trades = queryset.filter(bank=best_result[0])
                        if possible_trades.count() > 1 and possible_trades.filter(claims__isnull=False).exists():
                            possible_trades = possible_trades.filter(claims__isnull=False)
                        if (
                            possible_trades.count() >= 1
                        ):  # If count is greater than 1, we get exact match so we return either trades
                            queryset = possible_trades
            if (queryset.count() > 1) and (price := data.get("price", None)):
                if queryset.filter(price=price).count() == 1:
                    queryset = queryset.filter(price=price)
            if queryset.exists():
                # We try to filter by price as well
                trade = queryset.first()

                if queryset.count() == 1:
                    self.import_source.log += f"\nOne Trade found: {trade}"
                if queryset.count() > 1:
                    self.import_source.log += f"\nMultiple similar Trades found (returning first trade): {trade}"
                return trade
        self.import_source.log += "\nNo trade was successfully matched."

    def _get_history(self, history: Dict[str, Any]) -> models.QuerySet:
        trades = self.model.objects.filter(
            exclude_from_history=False,
            pending=False,
            transaction_subtype__in=[
                self.model.Type.SUBSCRIPTION,
                self.model.Type.REDEMPTION,
            ],  # we cannot exclude marked for deleted trade because otherwise they are never consider in the history
        )
        if transaction_date := history.get("transaction_date"):
            trades = trades.filter(transaction_date__lte=transaction_date)
        elif book_date := history.get("book_date"):
            trades = trades.filter(book_date__lte=book_date)
        if underlying_instrument_data := history.get("underlying_instrument"):
            if isinstance(underlying_instrument_data, dict):
                trades = trades.filter(
                    **{f"underlying_instrument__{k}": v for k, v in underlying_instrument_data.items()}
                )
            else:
                trades = trades.filter(underlying_instrument__id=underlying_instrument_data)

        elif "underlying_instruments" in history:
            trades = trades.filter(underlying_instrument__id__in=history["underlying_instruments"])
        else:
            raise ValueError("We cannot estimate history without at least the underlying instrument")
        return trades

    def _post_processing_objects(
        self,
        created_objs: list[models.Model],
        modified_objs: list[models.Model],
        unmodified_objs: list[models.Model],
    ):
        for instrument in set(
            map(lambda x: x.underlying_instrument, filter(lambda t: t.is_customer_trade, created_objs + modified_objs))
        ):
            if instrument.instrument_type.key == "product":
                update_outstanding_shares_as_task.delay(instrument.id)

    def _post_processing_updated_object(self, _object):
        if _object.marked_for_deletion:
            _object.marked_for_deletion = False
            _object.save()
            self.import_source.log += "\nMarked for deletion reverted"
        if _object.underlying_instrument.instrument_type == InstrumentType.PRODUCT:
            _object.link_to_internal_trade()

    def _post_processing_created_object(self, _object):
        self._post_processing_updated_object(_object)

    def _post_processing_history(self, history: models.QuerySet):
        self.import_source.log += "===================="
        self.import_source.log += "It was a historical import and the following Trades have to be deleted:"
        for trade in history.order_by("transaction_date"):
            self.import_source.log += (
                f"{trade.transaction_date:%d.%m.%Y}: {trade.shares} {trade.bank} ==> Marked for deletion"
            )
            trade.marked_for_deletion = True
            trade.save()



from datetime import date

import factory
from psycopg.types.range import DateRange

from wbportfolio.models import (
    InstrumentPortfolioThroughModel,
    Portfolio,
    PortfolioPortfolioThroughModel,
)


class PortfolioFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Portfolio

    name = factory.Sequence(lambda n: f"Portfolio {n}")
    currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyUSDFactory")
    is_manageable = True
    is_tracked = True
    is_lookthrough = False
    only_weighting = True
    invested_timespan = DateRange(date.min, date.max)

    @factory.post_generation
    def depends_on(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for portfolio in extracted:
                self.depends_on.add(portfolio)


class ModelPortfolioFactory(PortfolioFactory):
    @factory.post_generation
    def dependant_portfolios(self, create, extracted, **kwargs):
        PortfolioPortfolioThroughModel.objects.create(
            portfolio=PortfolioFactory.create(),
            dependency_portfolio=self,
            type=PortfolioPortfolioThroughModel.Type.MODEL,
        )


class InstrumentPortfolioThroughModelFactory(factory.django.DjangoModelFactory):
    instrument = factory.SubFactory("wbportfolio.factories.ProductFactory")
    portfolio = factory.SubFactory("wbportfolio.factories.PortfolioFactory")

    class Meta:
        model = InstrumentPortfolioThroughModel
        django_get_or_create = ("instrument",)



from wbcore.contrib.directory.viewsets.previews import EntryPreviewConfig
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field


class CompanyPreviewConfig(EntryPreviewConfig):
    def get_display(self) -> Display:
        fields = [
            [repeat_field(2, "computed_str")],
            ["primary_telephone", "primary_email"],
            [repeat_field(2, "primary_manager")],
            ["asset_under_management", "invested_assets_under_management_usd"],
            [repeat_field(2, "potential")],
        ]
        try:
            entry = self.view.get_object()
            if entry.profile_image:
                fields.insert(0, [repeat_field(2, "profile_image")])
        except Exception:
            pass

        return create_simple_display(fields)

    def get_buttons(self):
        buttons = super().get_buttons()
        buttons.append(
            bt.WidgetButton(key="asset_allocation_table", icon=WBIcon.DATA_LIST.icon, label="Asset Allocation"),
        )
        buttons.append(
            bt.WidgetButton(key="geographic_focus_table", icon=WBIcon.DATA_LIST.icon, label="Geographic Focus"),
        )
        return buttons



import numpy as np
import pandas as pd

from wbportfolio.models import Product, Trade


def get_portfolio(x):
    product = Product.objects.get(id=x)
    portfolio = product.primary_portfolio
    return portfolio.id


def parse(import_source):
    data = list()

    df = pd.read_excel(import_source.file.open(), usecols=[3, 4, 7, 10, 15, 16, 17, 18, 20, 21], engine="openpyxl")

    # Filter out all non transaction rows and remove RECORD_DESC col
    df = df[df["RECORD_DESC"].str.strip() == "TRANSACTION"]
    del df["RECORD_DESC"]

    # Convert timestamps to json conform date strings
    df["TRANS_DATE"] = df["TRANS_DATE"].apply(lambda x: x.date().strftime("%Y-%m-%d"))
    df["SETTL_DATE"] = df["SETTL_DATE"].apply(lambda x: x.date().strftime("%Y-%m-%d"))

    # Replace all nan values with empty str
    df[["REGISTER_FIRSTNAME", "CUST_REF"]] = df[["REGISTER_FIRSTNAME", "CUST_REF"]].replace(np.nan, "", regex=True)

    # Merge REGISTER_FIRSTNAME and CUST_REF and then remove both cols
    df["note"] = df["REGISTER_FIRSTNAME"] + df["CUST_REF"]
    del df["REGISTER_FIRSTNAME"]
    del df["CUST_REF"]

    # Create Product Mapping and apply to df
    product_mapping = {
        product["isin"]: product["id"]
        for product in Product.objects.filter(isin__in=df["ISIN"].unique()).values("id", "isin")
    }
    df["ISIN"] = df["ISIN"].apply(lambda x: product_mapping[x])

    # Rename Columns
    df.columns = [
        "bank",
        "underlying_instrument",
        "transaction_date",
        "value_date",
        "external_id",
        "price",
        "shares",
        "comment",
    ]
    df["transaction_subtype"] = df.shares.apply(lambda x: Trade.Type.REDEMPTION if x < 0 else Trade.Type.SUBSCRIPTION)
    df["portfolio"] = df["underlying_instrument"].apply(lambda x: get_portfolio(x))
    df["underlying_instrument"] = df["underlying_instrument"].apply(lambda x: {"id": x, "instrument_type": "product"})
    # Convert df to list of dicts
    data = list(df.T.to_dict().values())

    return {"data": data}



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ProductGroupEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:product_group-list",
            args=[],
            request=self.request,
        )



from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from django.db.models import Sum
from faker import Faker
from wbfdm.dataloaders.proxies import InstrumentDataloaderProxy
from wbfdm.enums import Financial

from wbportfolio.factories import AssetPositionFactory, PortfolioFactory
from wbportfolio.models import AssetPosition

from ..backends.portfolio_base import DataLoader

fake = Faker()


@pytest.mark.django_db
class TestPortfolioBaseDataloader:
    @pytest.fixture
    def dataloader(self, weekday):
        portfolio = PortfolioFactory.create()
        AssetPositionFactory.create_batch(5, portfolio=portfolio, date=weekday, is_estimated=False)
        dataloader = DataLoader(portfolio, weekday)
        base_df = dataloader.base_df
        columns = dataloader.columns_map.keys()
        dataloader.__dict__["market_capitalization_df"] = pd.DataFrame(
            np.random.randint(1e6, 1e9, size=(base_df.shape[0], len(columns))), index=base_df.index, columns=columns
        )
        dataloader.__dict__["ebit_usd_df"] = pd.DataFrame(
            np.random.randint(100, 100000, size=(base_df.shape[0], len(columns))), index=base_df.index, columns=columns
        )
        dataloader.__dict__["total_assets_usd_df"] = pd.DataFrame(
            np.random.randint(1e6, 1e9, size=(base_df.shape[0], len(columns))), index=base_df.index, columns=columns
        )
        dataloader.__dict__["liabilities_usd_df"] = pd.DataFrame(
            np.random.randint(1e6, 1e9, size=(base_df.shape[0], len(columns))), index=base_df.index, columns=columns
        )
        return dataloader

    @patch.object(InstrumentDataloaderProxy, "financials")
    def test__get_financial_df(self, mock_fct, dataloader: DataLoader):
        return_value = []
        base_df = dataloader.base_df
        for pivot_date in dataloader.pivot_dates:
            for instrument_id in base_df.index:
                return_value.append({"instrument_id": instrument_id, "year": pivot_date.year, "value": fake.pyfloat()})
        mock_fct.return_value = return_value
        res = dataloader._get_financial_df(Financial.EBIT)
        pd.testing.assert_index_equal(res.index, base_df.index)
        assert set(res.columns.tolist()) == set(dataloader.columns_map.keys())

    @patch.object(InstrumentDataloaderProxy, "financials")
    def test__get_financial_df_empty(self, mock_fct, dataloader: DataLoader):
        mock_fct.return_value = []
        with pytest.raises(ValueError):
            dataloader._get_financial_df(Financial.EBIT)

    def test_base_df(self, dataloader):
        base_df = dataloader.base_df
        total_portfolio_weight = AssetPosition.objects.filter(
            portfolio=dataloader.portfolio, date=dataloader.val_date
        ).aggregate(c=Sum("weighting"))["c"]
        for instrument_id, df in base_df.to_dict("index").items():
            for key, value in df.items():
                pos = AssetPosition.objects.get(
                    portfolio=dataloader.portfolio, date=dataloader.val_date, underlying_instrument=instrument_id
                )
                if key == "weighting":
                    assert float(getattr(pos, key) / total_portfolio_weight) == pytest.approx(value, abs=1e-6)
                else:
                    assert float(getattr(pos, key)) == value

    # Test portfolio metric computation
    def test_portfolio_ownership_df(self, dataloader):
        assert dataloader.portfolio_ownership_df.dropna().empty is False
        pd.testing.assert_frame_equal(
            dataloader.portfolio_ownership_df,
            (1.0 / dataloader.market_capitalization_df).multiply(dataloader.base_df.total_value_fx_usd, axis=0),
            check_exact=True,
        )

    def test_portfolio_ebit_usd_df(self, dataloader):
        assert dataloader.portfolio_ebit_usd_df.dropna().empty is False
        pd.testing.assert_frame_equal(
            dataloader.portfolio_ebit_usd_df,
            dataloader.portfolio_ownership_df * dataloader.ebit_usd_df,
            check_exact=True,
        )
        pd.testing.assert_series_equal(
            dataloader.portfolio_ebit_usd_df.aggregate, dataloader.portfolio_ebit_usd_df.sum(axis=0), check_exact=True
        )

    def test_portfolio_total_assets_usd_df(self, dataloader):
        assert dataloader.portfolio_total_assets_usd_df.dropna().empty is False
        pd.testing.assert_frame_equal(
            dataloader.portfolio_total_assets_usd_df,
            dataloader.portfolio_ownership_df * dataloader.total_assets_usd_df,
            check_exact=True,
        )
        pd.testing.assert_series_equal(
            dataloader.portfolio_total_assets_usd_df.aggregate,
            dataloader.portfolio_total_assets_usd_df.sum(axis=0),
            check_exact=True,
        )

    def test_portfolio_liabilities_usd_df(self, dataloader):
        assert dataloader.portfolio_liabilities_usd_df.dropna().empty is False
        pd.testing.assert_frame_equal(
            dataloader.portfolio_liabilities_usd_df,
            dataloader.portfolio_ownership_df * dataloader.liabilities_usd_df,
            check_exact=True,
        )
        pd.testing.assert_series_equal(
            dataloader.portfolio_liabilities_usd_df.aggregate,
            dataloader.portfolio_liabilities_usd_df.sum(axis=0),
            check_exact=True,
        )

    def test_portfolio_aggregation_capital_employed(self, dataloader):
        assert dataloader.portfolio_aggregation_capital_employed.dropna().empty is False
        pd.testing.assert_series_equal(
            dataloader.portfolio_aggregation_capital_employed,
            dataloader.portfolio_total_assets_usd_df.aggregate - dataloader.portfolio_liabilities_usd_df.aggregate,
            check_exact=True,
        )

    def test_portfolio_aggregation_roce(self, dataloader):
        assert dataloader.portfolio_aggregation_roce.dropna().empty is False
        pd.testing.assert_series_equal(
            dataloader.portfolio_aggregation_roce,
            dataloader.portfolio_ebit_usd_df.aggregate
            / dataloader.portfolio_aggregation_capital_employed.rolling(2).mean(),
            check_exact=True,
        )



# Generated by Django 5.0.14 on 2025-07-30 14:25

from django.db import migrations, models
from decimal import Decimal

class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0083_order_alter_trade_options_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderproposal',
            name='min_order_value',
            field=models.IntegerField(default=0, help_text='Minimum Order Value in the Portfolio currency', verbose_name='Minimum Order Value'),
        ),
        migrations.AddField(
            model_name='order',
            name='desired_target_weight',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'),
                                      help_text='Desired Target Weight (for compliance and audit)', max_digits=9,
                                      verbose_name='Desired Target Weight'),
        ),
    ]



from wbfdm.models import Instrument

from wbportfolio.models import Index, Product


def get_casted_portfolio_instrument(instrument: Instrument) -> Product | Index | None:
    try:
        return Product.objects.get(id=instrument.id)
    except Product.DoesNotExist:
        try:
            return Index.objects.get(id=instrument.id)
        except Index.DoesNotExist:
            return None



import random
from datetime import date, timedelta
from decimal import Decimal

import factory
from wbcore.contrib.currency.models import CurrencyFXRates
from wbfdm.factories import InstrumentPriceFactory

from wbportfolio.models import AssetPosition


def get_weekday(o):
    instrument = o.underlying_quote or o.underlying_instrument
    if instrument.id and instrument.assets.exists():
        latest_position = instrument.assets.latest("date").date
    else:
        latest_position = date(2020, 1, 1)

    latest_position += timedelta(days=1)
    while latest_position.weekday() > 4:
        latest_position += timedelta(days=1)

    return latest_position


class AssetPositionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = AssetPosition
        django_get_or_create = ("portfolio", "underlying_quote", "date", "portfolio_created")

    underlying_instrument = factory.SubFactory("wbfdm.factories.instruments.InstrumentFactory")
    underlying_quote = factory.LazyAttribute(lambda o: o.underlying_instrument)

    date = factory.LazyAttribute(lambda o: get_weekday(o))
    asset_valuation_date = factory.LazyAttribute(lambda o: o.date)

    initial_price = factory.Faker("pydecimal", min_value=100, max_value=120, right_digits=4)
    underlying_quote_price = factory.LazyAttribute(
        lambda o: InstrumentPriceFactory.create(
            instrument=o.underlying_instrument or o.underlying_quote,
            calculated=False,
            date=o.date,
            net_value=o.initial_price,
        )
    )

    initial_shares = factory.LazyAttribute(lambda o: Decimal(random.randint(10, 10000)))
    initial_currency_fx_rate = Decimal(1)
    weighting = factory.Faker("pydecimal", min_value=0.01, max_value=1.0, right_digits=4)

    portfolio = factory.SubFactory("wbportfolio.factories.portfolios.PortfolioFactory")
    portfolio_created = None

    currency = factory.LazyAttribute(lambda o: (o.underlying_instrument or o.underlying_quote).currency)
    currency_fx_rate_instrument_to_usd = factory.LazyAttribute(
        lambda o: CurrencyFXRates.objects.get_or_create(
            currency=o.currency, date=o.date, defaults={"value": Decimal(1.0)}
        )[0]
    )

    is_estimated = False

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        if (asset_valuation_date := kwargs.get("asset_valuation_date", None)) and (
            currency := kwargs.get("currency", None)
        ):
            kwargs["currency_fx_rate_instrument_to_usd"] = CurrencyFXRates.objects.get_or_create(
                date=asset_valuation_date, currency=currency, defaults={"value": 1.0}
            )[0]
        if (asset_valuation_date := kwargs.get("asset_valuation_date", None)) and (
            portfolio := kwargs.get("portfolio", None)
        ):
            kwargs["currency_fx_rate_portfolio_to_usd"] = CurrencyFXRates.objects.get_or_create(
                date=asset_valuation_date, currency=portfolio.currency, defaults={"value": 1.0}
            )[0]
        return super()._create(model_class, *args, **kwargs)

    # @factory.pre
    # def generate_currency_rates(self, create, extracted, **kwargs):
    #



from typing import Generator

import pandas as pd
from django.contrib.contenttypes.models import ContentType
from django.db import models
from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcore import serializers as wb_serializers
from wbfdm.models import InstrumentPrice

from wbportfolio.models import AssetPosition, Product


@register("Product Data Integrity", rule_group_key="portfolio")
class RuleBackend(backend.AbstractRuleBackend):
    OBJECT_FIELD_NAME: str = "product"

    class DataTypeChoices(models.TextChoices):
        INSTRUMENT_PRICE = "INSTRUMENT_PRICE", "Valuation"
        ASSET_POSITION = "ASSET_POSITION", "Asset Position"

    product: Product

    def is_passive_evaluation_valid(self) -> bool:
        return (
            self.product.is_active_at_date(self.evaluation_date)
            and AssetPosition.objects.filter(is_estimated=False, portfolio=self.product.portfolio).exists()
            and InstrumentPrice.objects.filter(calculated=False, instrument=self.product).exists()
        )

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Product)

    def _build_dto_args(self):
        return (self.product,)

    @classmethod
    def get_all_active_relationships(cls) -> models.QuerySet:
        return Product.active_objects.all()

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            data_type = wb_serializers.MultipleChoiceField(
                choices=cls.DataTypeChoices.choices,
                default=[cls.DataTypeChoices.ASSET_POSITION.value, cls.DataTypeChoices.INSTRUMENT_PRICE.value],
                label="Flags",
                help_text="Set the flags that will trigger the rule",
            )

            @classmethod
            def get_parameter_fields(cls):
                return ["data_type"]

        return RuleBackendSerializer

    def _process_dto(self, product: Product, **kwargs) -> Generator[backend.IncidentResult, None, None]:
        for lag_threshold in reversed(self.thresholds):
            numerical_range = lag_threshold.numerical_range

            last_asset_position_date = (
                AssetPosition.objects.filter(is_estimated=False, portfolio=product.portfolio).latest("date").date
            )
            last_instrument_price_date = (
                InstrumentPrice.objects.filter(calculated=False, instrument=product).latest("date").date
            )

            asset_position_lag = (
                len(
                    pd.date_range(
                        min([self.evaluation_date, last_asset_position_date]), self.evaluation_date, freq="B"
                    )
                )
                - 1
            )
            instrument_price_lag = (
                len(
                    pd.date_range(
                        min([self.evaluation_date, last_instrument_price_date]), self.evaluation_date, freq="B"
                    )
                )
                - 1
            )
            breached_data_type = []
            if (
                self.DataTypeChoices.ASSET_POSITION.value in self.data_type
                and asset_position_lag >= numerical_range[0]
                and asset_position_lag < numerical_range[1]
            ):
                breached_data_type.append("Asset Position")
            if (
                self.DataTypeChoices.INSTRUMENT_PRICE.value in self.data_type
                and instrument_price_lag >= numerical_range[0]
                and instrument_price_lag < numerical_range[1]
            ):
                breached_data_type.append("Valuation")

            if len(breached_data_type) > 0:
                max_lag = max([asset_position_lag, instrument_price_lag])
                yield backend.IncidentResult(
                    breached_object=product,
                    breached_object_repr=str(product),
                    breached_value=str(max_lag),
                    report_details={
                        "Lag": max_lag,
                        "Last Valuation Datapoint": f"{last_instrument_price_date:%d.%m.%Y}",
                        "Last Asset Position Datapoint": f"{last_asset_position_date:%d.%m.%Y}",
                        "Valuation Lag": instrument_price_lag,
                        "Asset Position Lag": asset_position_lag,
                    },
                    severity=lag_threshold.severity,
                )
                break



# Generated by Django 4.2.11 on 2024-04-05 12:06

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0017_instrument_instrument_security_index"),
        ("wbportfolio", "0057_alter_portfolio_preferred_instrument_classifications_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="PMSInstrument",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("wbfdm.instrument",),
        ),
    ]



from django.db.models import TextChoices

class ExecutionStatus(TextChoices):
    PENDING = "PENDING", "Pending"
    IN_DRAFT = "IN_DRAFT", "In Draft"
    COMPLETED = "COMPLETED", "Completed"
    CANCELLED = "CANCELLED", "Cancelled"
    REJECTED = "REJECTED", "Rejected"
    FAILED = "FAILED", "Failed"
    UNKNOWN = "UNKNOWN", "Unknown"

class RoutingException(Exception):
    def __init__(self, errors):
        # messages: a list of strings
        super().__init__()  # You can pass a summary to the base Exception
        self.errors = errors

    def __str__(self):
        return str(self.errors)
